/*
 *  tty.h -- 外部関数の宣言
 */

int opentty(char *name);		/* シリアルポートを開く */
void ttyinit(int fd, long speed);	/* 速度の設定 */
int ttyputc(int fd, char ch);		/* シリアルポートに文字を出力する */
int ttygetc(int fd);			/* シリアルポートから文字を入力する */
int closetty(int fd);			/* シリアルポートを閉じる */

#define FLOW_STOP 0
#define FLOW_START 1

#ifdef MSDOS
#ifdef RTSCTS
void flowctl(int start_stop);
#else
#define flowctl(x)	/* 何もしない */
#endif /* RTSCTS */
#else
#define flowctl(x)	/* 何もしない */
#endif /* MSDOS */
