int issjis(unsigned char ch);
int jtos(unsigned short ch);
int stoj(unsigned short ch);
int iskana(unsigned char ch);
int kanakana(unsigned char ch);
int etoj(unsigned short ch);
int jtoe(unsigned short ch);
int iseuc(unsigned char ch);
char *sjistojis(char *s);
char *jistosjis(char *s);
char *euctojis(char *s);
char *jistoeuc(char *s);
